"Python3 program to generate vocabulary from a corpus."

##  DOX2VOX :
##
##  Python3 vocabularizer :
##  Copyright 2010, 2015, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 12/06/2010
##  last revision : 13/11/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
import  time, random

##  if sys.platform.startswith("linu"):  ##  presume linux
    ##  sys.path.append(sys.argv[0])
##  else:  ##  presume windows
sys.path.append(sys.argv[0])
##  in case executed from somewhere else.

##  local libraries :
from globdefs import *
from voxglib import add2dic
from voxglib import numeric, numerize
from voxglib import parashow, showobj
from voxglib import tokline
from voxglib import toklines
from voxglib import punktran
from voxglib import getmode, request
from voxglib import bastats, geomean
from voxglib import corsums, corrval
from voxglib import ranx, midvalue


##  Global vars :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("filetail",".txt"),
    ("jobname",''),  ##  name to link related files
    ("wordonly",0,0,1),
    ("casefold",1,0,1),
    ("snipsize",115,4,8192),
    ("sortcol","sniprate"),  ##  default sort variable
    ("minfreq",3,1,),  ##  default is to skip hapax legomena
    ("topvocs",144,1,4096),  ##  number of vocab items to dump
    ("docpaths",''),  ##  folder(s) of input texts
    ("outpath",os.pardir+os.sep+"op"+os.sep),  ##  default o/p drive/path
    ("vocfile",''),  ##  vocabulary file, human-readable o/p
    ("vocdump",''),  ##  vocab file, machine-readable o/p
    ##  logfile ?
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval

##  parameters 2b requested from user :
Askables = ["jobname","atomize","casefold","docpaths","filetail","outpath",
           "snipsize","sortcol","topvocs","vocdump","vocfile","wordonly"]

Version = "1.4"
Mindocs = 2  ##  minimum number of docs in corpus
Minfreq = 3  ##  to exclude hapax & dislegomena
Toklimit = 64 * 1024  ##  large documents cause too much overhead (?inactive)
Feffmarx = ["\ufeff","\uffef"]  ##  to detect Windows faffing around with feff
Sortcols = "corprate corpsum  docrate sniprate textmean textmid"
Sortvars = Sortcols.split()
Headline = "`                     rank  corpfreq  docfreq  snipfreq  "
Headline = Headline + Sortcols
Maxsize = 20  ##  maximum token length !!
##  fixed parameters 2get started :
##  Docpath = "c:\\vocsoft\\samples\\handover\\western2"
##  Docpath = "c:\\vocsoft\\samples\\bottlabs\\docs\\"
##  Docpath = "c:\\tb\\feds\\fedpaps"
Vocfile = "voclist.txt"
Voctail = "_vocs.txt"
Voxtail = "_vocs.dat"


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    na = len(args)
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    print(progname,Version,dateline)
    print("command-line args. =",na)
    progpath = os.path.dirname(progname)  ##  directory from which program runs
    print("progpath :",progpath)
    ##  print(sys.path)
    whereat = os.getcwd()  ##  working directory
    print("working folder: ",whereat)
    ##  print(na,args)
    startime = time.time()

    ##  initialization :
    paralist = sorted(Paralist)
    paradat = Sack(progname)
    paradat.dateline = dateline
    paradat.progname = progname
    paradat.progpath = progpath
    paradat.whereat = whereat
    if na > 1:
        paradat.jobname = argv[1]
    else:
        paradat.jobname = ''
    ##  seek settings interally, then on file :
    paratab = init(paradat,paralist)
    ##  returns paralist info as look-up table.
    ##  establish o/p file defaults if none given :
    if not hasattr(paradat,"vocfile") or not paradat.vocfile:
        paradat.vocfile = paradat.jobname + Voctail
    if not hasattr(paradat,"vocdump") or not paradat.vocdump:
        paradat.vocdump = paradat.jobname + Voxtail    

    if not paradat.progpath.endswith(os.sep):
        paradat.progpath = paradat.progpath + os.sep
    if not paradat.whereat.endswith(os.sep):
        paradat.whereat = paradat.whereat + os.sep

    ##  obtain settings from user last, with prog & file inputs as defaults :
    paradat = prelude(paradat,paratab)
    if not paradat.docpaths.endswith(os.sep):
        paradat.docpaths = paradat.docpaths + os.sep
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
    if not paradat.progpath.endswith(os.sep):
        paradat.progpath = paradat.progpath + os.sep
    if not paradat.whereat.endswith(os.sep):
        paradat.whereat = paradat.whereat + os.sep
    paradat.pathlist = getpaths(paradat.docpaths)
    ##  parashow(paradat)

    ##  next read text files from selected folder(s) :
    paradat.doclist = []
    paradat.totsnips = 0
    for paradat.docpath in paradat.pathlist:
        doclist = readdocs(paradat)  ##  read from next folder
        paradat.doclist.extend(doclist)
    paradat.docs = len(paradat.doclist)
    print("texts read from",paradat.docpaths,'=',paradat.docs)
    ##  print(doclist[0].size)
    ##  print(doclist[0].toklist[:64])

    ##  then form token frequency list :
    freqtab = tokcount(paradat)
    for w in []:
        ##  ["bonk","fact","fail","free","hong","july","kong","year","zonk"]:
        nw = freqtab.get(w,0)
        print(w,nw)
    vocsize = len(freqtab)
    print("total word tokens =",paradat.tottoks)
    print("total vocabulary size =",vocsize)
    paradat.vocsize = vocsize

    ##  then form sorted frequency table :
    vocdict,voclist = vocsort(paradat,freqtab)
    print("tokens occuring at least",Minfreq,"times =",len(vocdict))
    for j in range(0):
        v4 = [round(v,4) for v in voclist[j][:-1]]
        print(j,v4,voclist[j][-1])

    ##  lastly write & dump vocabulary listing(s) :
    paradat.voutname = makefull(paradat.outpath,paradat.vocfile)
    paradat.dumpname = makefull(paradat.outpath,paradat.vocdump)
    ##  always to outpath ? -- probably safest.
    nout = vocwrite(paradat,freqtab,vocdict,voclist)
    print("wordform lines =",nout)
    print("vocabulary listing on",paradat.voutname)
    print("data values listed on",paradat.dumpname)
    
    ##  finale :
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,7),"seconds.")
    ##  secs since kickoff.
    
    return  0



##  GETPATHS :
def getpaths (docpaths,sepchar=','):
    "unpacks list of folders provided."

    parline = docpaths.lower().strip()
    pathbits = parline.split(sepchar)  ##  item separator defaults to comma
    ##  print(pathbits)
    pathlist = []
    for pb in pathbits:
        p = pb.strip()
        print(p)
        if os.path.isdir(p):
            pathlist.append(p)
        else:
            print("Warning :",p,"isn't an existing folder!")
        
    return  pathlist
    

##  PRELUDE :
def prelude (paradat,paratab):
    "gets user options."

    sortvars = Sortcols.split()  ##  sort column names
    sortvars.remove("corpsum")
    ##  print(sortvars)
    for j,a in enumerate(Askables):
        df = getattr(paradat,a)  ##  default to current value
        ##  print(j,a,df,numeric(df))
        if a == "sortcol":  ##  special case, handled below
            continue
        if a == "outpath":
            df = os.path.abspath(df)
        okay = False
        while not okay:
            pv = request(a,df)
            okay = parcheck(a,df,pv,paratab)
            if okay and numeric(df):
                pv = numerize(pv)
        if df != pv:
            print(a,'=',pv)
        setattr(paradat,a,pv)

    ##  sortcol as special case :
    a = "sortcol"
    pv = getmode("please choose "+a,sortvars)
    pv = sortvars[pv]
    ##  if pv == "corpsum":
        ##  pv = corprate"
    setattr(paradat,a,pv)
    print(a,'=',pv)

    return  paradat


##  PARCHECK :
def parcheck (a,df,pv,paratab):
    "does basic checking of parameter options given by user."

    okay = False
    if numeric(df):
        pv = numerize(pv)
        if pv == None:
            print(a,"needs numeric input!")
            return  False
        lo = paratab[a][2]
        if pv < lo:
            print(a,"requires value of",lo,"or more!")
            return  False
        if len(paratab[a]) > 3:
            hi = paratab[a][3]
            if pv > hi:
                print(a,"must not exceed",hi,'!')
                return  False
    elif a == "jobname":
        if not pv.isalnum():
            print("characters in",a,"must be alphanumeric!")
            return  False
        pv = pv.replace(' ','_')  ##  remove blanks
        ##  print(len(pv))
        return  (len(pv) > 0) and (len(pv) < 65)
    elif a in ["docpath","outpath"]:
        if os.path.isdir(pv):
            return  True
        ##  problem :
        print(pv,"must refer to an existing folder!")
        return  False


    return  True


##  INIT :
def init (paradat,paralist):
    "initialization routine; affects paradat."

    ##  first use internal defaults :
    paratab = {}  ##  will be lookup version of paralist
    for j,p in enumerate(paralist):
        ##  print(j,p)
        setattr(paradat,p[0],p[1])
        paratab[p[0]] = p  ##  saved 4 easy look-up
    ##  jobname is special, request from user early :
    okay = False
    while not okay:
        jn = request("please give jobname",paradat.jobname)
        if not jn.isalnum():
            print("characters in jobname must be alphanumeric!")
        elif len(jn) < 1:
            break  ##  leave default
        else:
            okay = True
            paradat.jobname = jn
    print(paradat.jobname,"to be used as jobname.")
    ##  parashow(paradat)

    tinf = Tokstuff(paradat.id)
    paradat.tokinfo = tinf  ##  tokenization info
    paradat.punxtab = punktran()  ##  default character equivalents

    ##  next overwrite from settings.txt if found :
    if not paradat.progpath.endswith(os.sep):
        paradat.progpath += os.sep
    sf = paradat.progpath + "settings.txt"
    ##  print(sf)
    ##  print(dir(paradat))  ##  better drop nonalphanumerics ?
    parinit(paradat,sf,paratab,dir(paradat))

    dirs = [paradat.whereat]
    ##  print(os.path.dirname(paradat.whereat))
    ##  print(os.path.split(paradat.whereat))
    ##  if len(bits) > 2:
        ##  d = os.sep.join(bits[:-2]) + os.sep + "parapath"
        ##  dirs.append(d)
    d = os.path.dirname(paradat.whereat)
    if d:
        dirs.append(d + os.sep + "parapath")
        dirs.append(d)  ##  parent last
    ##  print(dirs)
    ##  try2 read in order; exit on first success :
    for d in dirs:
        fn = makefull(d,paradat.jobname+".txt")
        ok = os.path.isfile(fn)
        if ok:  ##  file exists, try2 read from it
            parinit(paradat,fn,paratab,Askables)
            break

    return  paratab


##  PARINIT :
def parinit (paradat,pf,paratab,usables):
    "reads parameter values from given file, if present."

    if not os.path.isfile(pf):  ## belt & braces
        return
    fi = open(pf,'r',encoding="utf-8")
    setlines = fi.readlines()
    for setline in setlines:
        bits = setline.strip().split(maxsplit=1)
        ##  print(bits)
        p,pv = bits
        ##  print(p,pv)
        if not p.isalnum():
            continue  ##  valid para names will be alphanumeric
        ##  if hasattr(paradat,p) and p != "jobname":
        if p in usables and p != "jobname":  ##  jobname must stick
            ##  reset value if valid :
            dv = getattr(paradat,p)  ##  current value is default
            if numeric(dv):
                pv = numerize(pv)
                if pv == None:
                    pass  ##  just ignore
                else:
                    lo = paratab[p][2]
                    if pv < lo:
                        pv = lo
                    if len(paratab[p]) > 3:
                        hi = paratab[p][3]
                    if pv > hi:
                        pv = hi
                setattr(paradat,p,pv)
            else:  ##  string value
                if p == "sortcol":
                    if pv in Sortvars:
                        setattr(paradat,p,pv)
                    else:
                        pass
                else:  ##  no other checks
                    setattr(paradat,p,pv)
    fi.close()
    ##  parashow(paradat)

    return


##  VOCWRITE :
def vocwrite (paradat,freqtab,vocdict,voclist):
    "writes vocab listing onto _vocs file."

    nout = 0
    fo = open(paradat.voutname,'w',encoding="utf-8")
    f2 = open(paradat.dumpname,'w',encoding="utf8")
    ##  heading :
    fo.write(Headline+NL)
    h = Headline.replace('`',"wordform")
    h = h.split()
    f2.write(Htab.join(h) + NL)
    corpsum = 0.0
    for j,v in enumerate(voclist[:paradat.topvocs]):
        t = v[-1]  ##  token held in last spot
        ##  v[0] can be ignored (repeated sortkey)
        frex = vocdict[t]
        corpsum += v[1]
        lout = "{:20s}  {:4d} {:9d} {:8d} {:9d}" \
               .format(t[:Maxsize],j+1,frex[0],frex[1],frex[2])
        lout += "  {:8.2f} {:7.2f} {:8.2f} {:8.2f} {:8.2f} {:7.2f}" \
                .format(v[1],corpsum,v[2],v[3],v[4],v[5])
        print(lout,file=fo)
        nout += 1
        bits = lout.split()
        f2.write(Htab.join(bits)+NL)  ##  also dump in R-style format

    fo.write(NL)
    parashow(paradat,fo)

    fo.close() ; f2.close()

    return  nout


##  VOCSORT :
def vocsort (paradat,freqtab):
    "returns extended vocab table & voclist for sorting."

    vocdict = {}
    voclist = []
    ##  sp = sortpos(paradat.sortcol)
    sv = Sortvars
    sv.remove("corpsum")
    sp = sv.index(paradat.sortcol)
    ##  skip over corpsum
    ##  print(paradat.sortcol,sp)
    for t in freqtab:  ##  for each vocabulary item
        freq = freqtab[t]
        if freq < Minfreq:
            continue  ##  too rare to bother with
        corprate = 100.0 * freq / paradat.tottoks  ##  all rates as percentages
        ratelist = []
        snipfreq = 0
        docfreq = 0
        for d in paradat.doclist:
            f = d.freqtab.get(t,0)
            rate = f / max(1,d.size)
            ratelist.append(rate)
            if f > 0:
                docfreq += 1
            for s in d.snipsets:
                if t in s:
                    snipfreq += 1
        docrate = 100.0 * docfreq / paradat.docs
        sniprate = 100.0 * snipfreq / paradat.totsnips
        textmean = 100.0 * sum(ratelist) / paradat.docs
        textmid = midvalue(ratelist) * 100.0
        vals = [corprate,docrate,sniprate,textmean,textmid,t]
        ##  vals = [vals[sp]] + vals  ##  sort value at front
        vals.insert(0,vals[sp])
        basedat = [freq,docfreq,snipfreq]
        vocdict[t] = basedat
        voclist.append(vals)

    voclist.sort(reverse=True)

    return  vocdict,voclist


##  ADDSNIPS :
def addsnips (d,snipsize):
    "adds snippet data to text d if not already present."

    d.snippets = 0
    d.snipsets = []  ##  list of sets
    minsnip = int(round(snipsize / Gold))
    for p in range(0,d.size,snipsize):
        slab = d.toklist[p:p+snipsize]
        if len(slab) >= minsnip or p==0:
            d.snipsets.append(set(slab))
            d.snippets += 1

    return  ##  side-effect on d.


##  TOKCOUNT :
def tokcount (paradat):
    "makes freq tables from toklists in doclist."

    mainfrex = {}
    paradat.tottoks = 0
    for d in paradat.doclist:
        subtab = getfrex(d.toklist)
        d.freqtab = subtab  ##  save in case useful later processing
        add2dic(mainfrex,subtab)  ##  acculumulate overall dictionary
        paradat.tottoks += d.size
        if "zonq" in subtab:
            print(d.name,subtab["zonq"])

    return  mainfrex        
        

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  READDOCS :
def readdocs (paradat):
    "reads in text docs from selected folder; affects paradat."

    filelist = os.listdir(paradat.docpath)
    nf = len(filelist)
    print("files found on",paradat.docpath,'=',nf)
    if not paradat.docpath.endswith(os.sep):
        paradat.docpath += os.sep  ##  redundant ?
    doclist = []
    totsnips = 0
    ##  input loop :
    for fn in filelist:
        okay = (paradat.filetail == '') or \
               fn.lower().endswith(paradat.filetail.lower())
        if okay:
            filename = makefull(paradat.docpath,fn)
            rawlines = try2read(filename)
            d = Sack(filename)
            d.name = fn ; d.prepath = paradat.docpath
            d.rawlines = rawlines
            if paradat.atomize:
                d.toklist = toklines(rawlines,paradat)
            else:  ##  just split tokens at whitespace
                d.toklist = []
                for r in rawlines:
                    d.toklist.append(r.split())
            ##  d.toklist = d.toklist[:Toklimit]
            d.size = len(d.toklist)
            addsnips(d,paradat.snipsize)  ##  keep snippet sets
            totsnips += d.snippets
            doclist.append(d)
    paradat.totsnips += totsnips

    return  doclist


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  TRY2READ :
def try2read (f,codelist=["utf8","windows-1252","latin1"]):
    "tries more than once to read text file by guessing (?Windows) encodings."
    ##  utf8 is normally first choice.
    
    xxxx = None
    for ec in codelist:
        try:
            textfile = open(f,'r',encoding=ec)
            rawlines = textfile.readlines()
            textfile.close()
            return  rawlines
        except (UnicodeError,ValueError,IOError) as xxxx:
            rawlines = []

    ##  back-stop procedure is to replace error bytes (sort-of) :
    textfile = open(f,'r',encoding="utf8",errors="surrogateescape")
    rawlines = textfile.readlines()
    textfile.close()
    
    return  rawlines



            
##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
