"Python3 program to make htab-delimited data file from vocabulary listing."

##  VOX2DAT :
##
##  Python3 program to compute data grid of docvars from vocabulary list :
##  Copyright 2010, 2015, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 12/06/2010
##  last revision : 13/11/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
import  math, os, string, sys
import  random, time

sys.path.append(sys.argv[0])
##  in case executed from somewhere else.

##  local libraries :
from globdefs import *
from voxglib import add2dic
from voxglib import numeric, numerize
from voxglib import parashow, showobj
from voxglib import tokline
from voxglib import toklines
from voxglib import punktran
from voxglib import getmode, request
from voxglib import bastats, geomean
from voxglib import corsums, corrval
from voxglib import ranx, midvalue
from voxglib import diversim
from dox2vox import getpaths


##  Global vars :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("filetail",".txt"),
    ("jobname",''),  ##  name to link related files
    ("wordonly",0,0,1),
    ("casefold",1,0,1),
    ("snipsize",115,4,8192),  ##  snippet size
    ("topvocs",144,1,4096),  ##  number of vocab items to dump
    ("folders",''),  ##  folder(s) of input texts
    ("outpath",os.pardir+os.sep+"op"+os.sep),  ##  default o/p drive/path
    ("vocdump",''),  ##  potential vocab file made by dox2vox.py
    ("voxfile",''),  ##  vocabulary file, input
    ("datfile",''),  ##  o/p in R-type format
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval

##  parameters 2b requested from user :
Askables = ["jobname","atomize","casefold","datfile","filetail","folders",
            "outpath","snipsize","topvocs","voxfile","wordonly"]
        ##  + classlab ?

Version = "1.4"
Mindocs = 2  ##  minimum number of docs in corpus
Minfreq = 3  ##  to exclude hapax & dislegomena
Toklimit = 64 * 1024  ##  large documents cause too much overhead ?
Feffmarx = ["\ufeff","\uffef"]  ##  to detect Windows faffing around with feff
Headline = "prepath  textname  filenum  totchars  tottoks  totvocs"
Heading = Headline.split()  ##  list version
Maxsize = 20  ##  maximum token length !!
Voctail = "_vocs.txt"
Voxtail = "_vocs.dat"
Dattail = "_vars.dat"


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    na = len(args)
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    print(progname,Version,dateline)
    print("command-line args. =",na)
    progpath = os.path.dirname(progname)  ##  directory from which program runs
    print("progpath :",progpath)
    ##  print(sys.path)
    whereat = os.getcwd()  ##  working directory
    print("working folder: ",whereat)
    ##  print(na,args)
    startime = time.time()

    ##  initialization :
    paralist = sorted(Paralist)
    paradat = Sack(progname)
    paradat.dateline = dateline
    paradat.progname = progname
    paradat.progpath = progpath
    paradat.whereat = whereat
    if na > 1:
        paradat.jobname = argv[1]
    else:
        paradat.jobname = ''

    ##  seek settings internally, then on file :
    paratab = init(paradat,paralist)
    ##  returns paralist info as look-up table.
    if not paradat.progpath.endswith(os.sep):
        paradat.progpath = paradat.progpath + os.sep
    if not paradat.whereat.endswith(os.sep):
        paradat.whereat = paradat.whereat + os.sep

    ##  parashow(paradat)

    ##  revise settings from user input, with prog & file inputs as defaults :
    paradat = prelude(paradat,paratab)

    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
    if not paradat.progpath.endswith(os.sep):
        paradat.progpath = paradat.progpath + os.sep
    if not paradat.whereat.endswith(os.sep):
        paradat.whereat = paradat.whereat + os.sep
    paradat.pathlist = getpaths(paradat.folders)
    ##  parashow(paradat)

    ##  read input vocabulary :
    ni = readvocs(paradat)
    vf = paradat.voxfile
    if ni < 1:
        vf = "file"
    print("vocabulary items read from",vf,'=',ni)
    ##  print(paradat.voclist)
    
    ##  next read text files from selected folder :
    paradat.doclist = []
    paradat.totsnips = 0
    for paradat.docpath in paradat.pathlist:
        doclist = readdocs(paradat)
        paradat.doclist.extend(doclist)
    ##  paradat.docpath = ''  ##  clunky
    paradat.docs = len(paradat.doclist)
    print("texts read from",paradat.folders,'=',paradat.docs)
    ##  print(doclist[0].size)
    ##  print(doclist[0].toklist[:64])
    ##  print(doclist[-1].size)
    ##  print(doclist[-1].toklist[:64])
    grid = makegrid(paradat)
    ##  for g in grid[:4]:
        ##  print(g[:19])

    ##  amend heading :
    for v in paradat.voclist:
        if v.isalpha():
            if len(v) >= 4:
                Heading.append(v)
            else:
                Heading.append(v+'_')
        else:
            Heading.append("v_"+v)  ## safer for spss ?
    
    paradat.dumpname = makefull(paradat.outpath,paradat.datfile)
    nout = griddump(paradat,grid)
    print("output lines =",nout)
    print("data values listed on",paradat.dumpname)
    paradat.pardump = makefull(paradat.outpath,"vox2dat_pars.txt")
    fo = open(paradat.pardump,'w',encoding="utf-8")
    parashow(paradat,fo)
    fo.close()
    
    ##  finale :
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,7),"seconds.")
    ##  secs since kickoff.
    
    return  0



##  MAKEGRID :
def makegrid (paradat):
    "makes data matrix from text docs & vocab file."

    grid = []
    for j,d in enumerate(paradat.doclist):
        nc = 0
        for r in d.rawlines:
            rawline = r.strip()
            nc += len(rawline)  ##  + 1 ??
        d.totchars = nc
        ##  next compute relative token frex :
        d.freqtab = getfrex(d.toklist)  ##  readdocs doesn't compute this
        v = len(d.freqtab)
        datline = [os.path.dirname(d.id),d.name,j+1,d.totchars,d.size,v]        
        ratelist = []
        for j,t in enumerate(paradat.voclist):
            f = d.freqtab.get(t,0)
            rate = f / max(1,d.size)
            ratelist.append(100.0*rate)
        ##  compute diversity indices :
        divlist = divrates(paradat,d)
        datline += divlist + ratelist
        grid.append(datline)

    ##  amend heading; must be coordinated with divlist :
    Heading.extend(["bw","diversim","haprate","herdanc","hr","v2overv"])
    Heading.extend(["sniphaps","shsd","snipttr","stsd"])

    return  grid


##  DIVRATES :
def divrates (paradat,d):
    "computes diversity indices of text d."

    ##  overall indices :
    v = len(d.freqtab)
    bw = v ** (-0.169)
    bw = d.size ** bw  ##  Brunet's W
    haps = 0 ; v2 = 0
    simpvec = []
    for t in d.freqtab:
        f = d.freqtab[t]
        if f == 1:
            haps += 1
        elif f == 2:
            v2 += 1
            simpvec.append(f)
        else:
            simpvec.append(f)
    ##  dugastd = math.log(v) / math.log(math.log(d.size))
    ##  dugast d is rubbish, removed !
    ds = diversim(simpvec)  ## + [haps])  ##  simpson's but without haps
    haprate = haps / v
    herdanc = math.log(v) / math.log(d.size)
    hr = 100.0*math.log(d.size) / (1.0 - haps / (v+0.5))
    v2overv = v2 / v
    ##  snippet-based measures :
    ttrlist = []
    for s in d.snipsets:
        ttrlist.append(100*len(s) / paradat.snipsize)
    bs = bastats(ttrlist)
    sttr = bs.mean
    if d.snippets > 1:
        s = bs.std2
    else:
        s = 0.0
    bs = bastats(d.haplist)
    sniphaps = 100.0 * bs.mean
    if d.snippets > 1:
        sd = bs.std2 * 100.0
    else:
        sd = 0.0

    ##  append vars  :
    divlist = [bw,ds,haprate,herdanc,hr,v2overv]
    divlist.extend([sniphaps,sd,sttr,s])

    return  divlist
    ##  must amend Heading (in makegrid) as measures are added.


##  READVOCS :
def readvocs (paradat):
    "reads in vocabulary file; alters paradat."

    paradat.voclist = []
    if hasattr(paradat,"voxfile") and os.path.isfile(paradat.voxfile):
        fn = paradat.voxfile  ##  already gives full path
    else:
        fn = makefull(paradat.outpath,paradat.voxfile)
    if not os.path.isfile(fn):
        print("vocabulary file",fn,"not found!")
        return  0

    ##  read _vocs.dat file :
    fi = open(fn,'r',encoding="utf8")
    voclines = fi.readlines()
    ##  print(len(voclines))
    fi.close()
    voclist = []
    ni = 0
    for vocline in voclines[1:]:  ##  line 1 is heading
        bits = vocline.split()  ##  Htab)
        t = bits[0].strip()
        if t:  ##  skip blanks
            voclist.append(t)
            ni += 1
            ##  print(ni,t)
        if len(voclist) >= paradat.topvocs:
            break  ##  only use number sought

    paradat.voclist = voclist
    return  len(voclist)

        
##  PRELUDE :
def prelude (paradat,paratab):
    "gets user options."

    for j,a in enumerate(Askables):
        df = getattr(paradat,a)  ##  default to current value
        ##  print(j,a,df,numeric(df))
        if a == "jobname":  ##  special case
            if df:  ##  already given
                continue
        if df and a in ["outpath","textpath"]:
            df = os.path.abspath(df)
        okay = False
        while not okay:
            pv = request(a,df)
            okay = parcheck(a,df,pv,paratab)
            if okay and numeric(df):
                pv = numerize(pv)
        if df != pv:
            print(a,'=',pv)
        setattr(paradat,a,pv)

    return  paradat


##  PARCHECK :
def parcheck (a,df,pv,paratab):
    "does basic checking of parameter options given by user."

    okay = False
    if numeric(df):
        pv = numerize(pv)
        if pv == None:
            print(a,"needs numeric input!")
            return  False
        lo = paratab[a][2]
        if pv < lo:
            print(a,"requires value of",lo,"or more!")
            return  False
        if len(paratab[a]) > 3:
            hi = paratab[a][3]
            if pv > hi:
                print(a,"must not exceed",hi,'!')
                return  False
    elif a == "jobname":
        if not pv.isalnum():
            print("characters in",a,"must be alphanumeric!")
            return  False
        pv = pv.replace(' ','_')  ##  remove blanks
        ##  print(len(pv))
        return  (len(pv) > 0) and (len(pv) < 65)
    elif a in ["outpath","textpath"]:
        if os.path.isdir(pv):
            return  True
        ##  problem :
        print(pv,"must refer to an existing folder!")
        return  False

    return  True


##  INIT :
def init (paradat,paralist):
    "initialization routine; affects paradat."

    ##  first use internal defaults :
    paratab = {}  ##  will be lookup version of paralist
    for j,p in enumerate(paralist):
        ##  print(j,p)
        setattr(paradat,p[0],p[1])
        paratab[p[0]] = p  ##  saved 4 easy look-up
    ##  jobname is special, request from user :
    if '.' in paradat.jobname:
        bits = paradat.jobname.split('.')
        paradat.jobname = bits[0]
    okay = False
    while not okay:
        jn = request("please give jobname",paradat.jobname)
        if not jn.isalnum():
            print("characters in jobname must be alphanumeric!")
        elif len(jn) < 1:
            break  ##  leave default
        else:
            okay = True
            paradat.jobname = jn
    ##  also check that _vocs.dat file exists ??
    print(paradat.jobname,"to be used as jobname.")
    ##  parashow(paradat)

    tinf = Tokstuff(paradat.id)
    paradat.tokinfo = tinf  ##  tokenization info
    paradat.punxtab = punktran()  ##  default character equivalents

    ##  next overwrite from settings.txt if found :
    if not paradat.progpath.endswith(os.sep):
        paradat.progpath += os.sep
    sf = paradat.progpath + "settings.txt"
    ##  print(sf)
    ##  print(dir(paradat))  ##  better drop nonalphanumerics ?
    parinit(paradat,sf,paratab,dir(paradat))

    dirs = [paradat.whereat]
    d = os.path.dirname(paradat.whereat)
    if d:
        dirs.append(d + os.sep + "parapath")
        dirs.append(d)  ##  parent last
    print(dirs)
    ##  try2 read in order; exit on first success :
    for d in dirs:
        fn = makefull(d,paradat.jobname+".txt")
        ok = os.path.isfile(fn)
        ##  print('*',fn)
        if ok:  ##  file exists, try2 read from it
            parinit(paradat,fn,paratab,Askables+["vocdump"])
            break

    ##  form file names from jobname :
    if hasattr(paradat,"voxfile") and os.path.isfile(paradat.voxfile):
        pass  ##  voxfile okay as given
    elif hasattr(paradat,"vocdump") and os.path.isfile(paradat.vocdump):
        paradat.voxfile = paradat.vocdump  ##  vocdump second choice, if given
        ##  does it need 2b full path spec ??
    else:
        paradat.voxfile = paradat.jobname + Voxtail  ##  make up default name
    if not paradat.datfile:
        paradat.datfile = paradat.jobname + Dattail

    ##  check outpath :
    op = getattr(paradat,"outpath",'')
    if op:
        op = os.path.abspath(op)
        if os.path.isdir(op):
            setattr(paradat,"outpath",op)
        else:
            setattr(paradat,"outpath",'')  ##  don't keep nonexistent path
        
    return  paratab


##  PARINIT :
def parinit (paradat,pf,paratab,usables):
    "reads parameter values from given file, if present."

    if not os.path.isfile(pf):  ## belt & braces
        return
    fi = open(pf,'r',encoding="utf-8")
    setlines = fi.readlines()
    for setline in setlines:
        bits = setline.strip().split(maxsplit=1)
        ##  print(bits)
        p,pv = bits
        ##  print(p,pv)
        if not p.isalnum():
            continue  ##  valid para names will be alphanumeric
        ##  if hasattr(paradat,p) and p != "jobname":
        if p in usables and p != "jobname":  ##  jobname must stick
            ##  reset value if valid :
            dv = getattr(paradat,p)  ##  current value is default
            if numeric(dv):
                pv = numerize(pv)
                if pv == None:
                    pass  ##  just ignore
                else:
                    lo = paratab[p][2]
                    if pv < lo:
                        pv = lo
                    if len(paratab[p]) > 3:
                        hi = paratab[p][3]
                    if pv > hi:
                        pv = hi
                setattr(paradat,p,pv)
            else:  ##  string value
                if p == "outpath":
                    pv = os.path.abspath(pv)  ##  expand
                    if not os.path.isdir(pv):
                        pv = ''  ##  don't allow nonexistent path
                    setattr(paradat,p,pv)
                else:  ##  no other checks
                    setattr(paradat,p,pv)
    fi.close()
    ##  print("leaving parinit....")
    ##  parashow(paradat)

    return


##  GRIDDUMP :
def griddump (paradat,grid):
    "writes data matrix onto _vars file."

    nout = 0
    ##  if exists, open in append mode ?
    fo = open(paradat.dumpname,'w',encoding="utf8")
    ##  heading :
    ##  h = Headline.split()
    fo.write(Htab.join(Heading) + NL)
    for j,gridline in enumerate(grid):
        bits = []
        for item in gridline:
            if isinstance(item,float):
                bits.append(str(round(item,7)))
            else:
                bits.append(str(item))
        
        fo.write(Htab.join(bits)+NL)  ##  also dump in R-style format
        nout += 1

    ##  fo.write(NL)
    fo.close()

    return  nout


##  ADDSNIPS :
def addsnips (d,snipsize):
    "adds snippet data to text d if not already present."

    d.snippets = 0
    d.snipsets = []  ##  list of sets
    d.haplist = []  ##  list of hapax rates per snippet
    minsnip = int(round(snipsize / Gold))
    for p in range(0,d.size,snipsize):
        slab = d.toklist[p:p+snipsize]
        if len(slab) >= minsnip or p==0:
            d.snipsets.append(set(slab))
            d.snippets += 1
            slabfrex = getfrex(slab)
            haps = 0
            for t in slabfrex:
                if slabfrex[t] == 1:
                    haps += 1
            haprate = haps / len(slab)
            d.haplist.append(haprate)
            
    return  ##  side-effect on d.


##  TOKCOUNT :
def tokcount (paradat):
    "makes freq tables from toklists in doclist."

    mainfrex = {}
    paradat.tottoks = 0
    for d in paradat.doclist:
        subtab = getfrex(d.toklist)
        d.freqtab = subtab  ##  save in case useful later processing
        add2dic(mainfrex,subtab)  ##  acculumulate overall dictionary
        paradat.tottoks += d.size
        if "zonq" in subtab:
            print(d.name,subtab["zonq"])

    return  mainfrex        
        

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  READDOCS :
def readdocs (paradat):
    "reads in text docs from selected folder."

    filelist = os.listdir(paradat.docpath)
    nf = len(filelist)
    print("files found on",paradat.docpath,'=',nf)

    doclist = []
    totsnips = 0
    ##  input loop :
    for fn in filelist:
        okay = (paradat.filetail == '') or \
               fn.lower().endswith(paradat.filetail.lower())
        if okay:
            filename = makefull(paradat.docpath,fn)
            rawlines = try2read(filename)
            d = Sack(filename)
            d.name = fn
            d.rawlines = rawlines
            if paradat.atomize:
                d.toklist = toklines(rawlines,paradat)
            else:  ##  just split tokens at whitespace
                d.toklist = []
                for r in rawlines:
                    d.toklist.append(r.split())
            ##  d.toklist = d.toklist[:Toklimit]
            d.size = len(d.toklist)
            if getattr(paradat,"snipsize",0) > 0:
                addsnips(d,paradat.snipsize)  ##  keep snippet sets
                totsnips += d.snippets
            doclist.append(d)
    paradat.totsnips += totsnips

    return  doclist


##  MAKEFULL :
def makefull (prepath,filename):
    "joins prepath to filename, with separator if missing."

    if prepath.endswith(os.sep):
        fullname = prepath + filename
    else:  ##  insert separator
        fullname = prepath + os.sep + filename

    return  fullname


##  TRY2READ :
def try2read (f,codelist=["utf8","windows-1252","latin1"]):
    "tries more than once to read text file by guessing (?Windows) encodings."
    ##  utf8 is normally first choice.
    
    xxxx = None
    for ec in codelist:
        try:
            textfile = open(f,'r',encoding=ec)
            rawlines = textfile.readlines()
            textfile.close()
            return  rawlines
        except (UnicodeError,ValueError,IOError) as xxxx:
            rawlines = []

    ##  back-stop procedure is to replace error bytes (sort-of) :
    textfile = open(f,'r',encoding="utf8",errors="surrogateescape")
    rawlines = textfile.readlines()
    textfile.close()
    
    return  rawlines



            
##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
