"Python3 utility routines for vocsoft programs, by R.S. Forsyth."

##  VOCSLIB :
##
##  vocsoft utility routines :
##  Copyright 2012, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 28/08/2015
##  last revision : 14/11/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
	
import  math, os, string, sys

from globdefs import *

Htab = '\n'
Maxtoke = 32  ##  maximum markup symbol size
##  Difficult punctuation signs (for punktran()) :
Punxlist = [-39,96,180,700,2036,8216,8217,8218,8219,8242]  ##  apoplectic !
Punxlist = Punxlist + [-45,173,0x2010,0x2011,0xfe63,0xfe0d]  ##  dash it all !
Punxlist = Punxlist + [-34,171,187,8220,8221,8222,12291]  ##  don't quote me !
Punxlist = Punxlist + [-126,0x2012,0x2013,0x2014,0x2015,0xfe58]  ## very dashing !
##  negatives are good old ascii substitutes.
Lowchin = 0x4e00  ##  lo point of unicode CJKblock (19968)
Highchin = 0x9fff  ##  hi point of unicode CJKblock (40907)


##  ADD2DIC :
def add2dic (largedic,smalldic):
    "augments contents of largedic with context of smalldic, both freq tables."

    ##  dk = list(smalldic.keys())
    for k in smalldic:
        f = smalldic[k]
        freq = largedic.get(k,0)
        largedic[k] = f + freq


##  NUMERIC :
def numeric (x):
    "tests whether x is int or float."

    if isinstance(x,type(9)) or isinstance(x,type(9.9)):
        return  True

    return  False


##  NUMERIZE :
def numerize(s):
    "Converts s to a numeric value if possible, else returns None."

    try:
        v = int(s)
    except ValueError:
        try:
            v = float(s)
        except ValueError:
            v = None  ## means s doesn't look like a number
    return v


##  PARASHOW :
def parashow (paradat,fo=sys.stdout,heading="Parameter settings :"):
    "displays settings in Sack() object paradat on file fo."

    showobj(paradat,fo,heading)

    return


##  PARADUMP :
def paradump (paradat,fo=sys.stdout,heading="Parameter settings :"):
    "dumps settings in Sack() object paradat onto file fo."

    dumpobj(paradat,fo,heading)

    return


##  DUMPOBJ :
def dumpobj (item,fo=sys.stdout,heading=''):
    "generic object display routine."

    entries = list(item.__dict__.items())
    entries.sort()
    if heading:
        fo.write(heading + NL)
    for e in entries:
        fo.write(str(e) + NL)
    fo.write(NL)

    return


##  SHOWOBJ :
def showobj (item,listing=sys.stdout,heading=''):
    "writes printable values of item (usually paradat) onto file."

    if heading:
        listing.write(heading + NL)
    parlist = sorted(dir(item))
    for p in parlist:
        if p.startswith("__"):  ##  private
            continue
        v = getattr(item,p)
        pout = ''
        if isinstance(v,int):
            pout = str(v)
        elif isinstance(v,float):
            pout = str(round(v,4))
        elif isinstance(v,str):
            pout = v
        elif isinstance(v,(dict,list,set,tuple)):
            if len(v) < 12:
                pout = str(v)
            else:
                pout = str(str(type(v))+" of "+str(len(v))+" items.")
        else:
            fd = dir(v)
            if "mode" in fd and ("name" in fd):  ##  must be a file ?
                pout = "~ " + v.name
        if pout:
            print(p.ljust(9),pout,file=listing)

    print('',file=listing)
    return


##  TOKLINE :
def tokline (rawline,paradat):
    "Tokenizes a single line of text; tinf contains char-type info."

    ##  return  string.split(line)
    word = mark = numb = punc = 0
    chin = 0
    tokelist = []
    toke = ''
    tinf = paradat.tokinfo  ##  should be set in globdefs.py
    tt = getattr(paradat,"punxtab",{})
    if tt:  ##  convert funny punctuation characters
        line = rawline.translate(tt)
    else:
        line = rawline
    for ch in line:
        ##  first check continuation :
        if mark:  ##  inside a mark-up symbol
            if ch == tinf.markshut or len(toke) > Maxtoke:
                mark = 0
                tokelist.append(toke + ch)
                toke = ''
            else:  toke = toke + ch
        elif word:  ##  inside a word
            ##  altered 18/08/2012 to do crude 1-char ChinJap tokenizing :
            if (ch.isalpha() or ch in tinf.wordcont) and \
               (ord(ch) < Lowchin or ord(ch) > Highchin):
                toke = toke + ch
            else:
                word = 0
                if toke[-1] == "'":
                    toke = toke[:-1]  ##  removes trailing quote
                tokelist.append(toke) ; toke = ''
        elif numb:
            if ch in tinf.numcont:
                toke = toke + ch
            else:
                numb = 0
                tokelist.append(toke) ; toke = ''
        elif punc:
            if ch in tinf.puncmarx:
                toke = toke + ch
            else:
                punc = 0
                tokelist.append(toke) ; toke = ''
        elif chin:
            tokelist.append(toke)
            chin = 0 ; toke = ''  ##  chin bloc simply split into chars
        else:
            pass  ##  ignore everything else

        ##  next check initiation :
        if chin + mark + word + numb + punc == 0:  ##  outside token
            if ch == tinf.markopen:
                toke = ch ; mark = 1
            elif ord(ch) >= Lowchin and ord(ch) <= Highchin:
                toke = ch ; chin = 1
            elif ch.isalpha():
                toke = ch ; word = 1
            elif ch in tinf.numstart:
                toke = ch ; numb = 1
            elif ch in tinf.puncmarx:
                toke = ch ; punc = 1
            else:
                pass  ##  ignores the rest
        
    return tokelist
    ##  no size limit to tokens here, but may be imposed afterwards.


##  TOKLINES :
def toklines (rawlines,paradat):
    "takes list of text lines as produced by readlines(), returns a list of tokens."

    toklist = []
    folding = 0
    if hasattr(paradat,"casefold"):
        folding = paradat.casefold
    if hasattr(paradat,"foldcase"):
        folding = paradat.foldcase
    for r in rawlines:
        rx = r.strip()
        if folding:
            rx = rx.lower()
        if paradat.atomize:
            toks = tokline(rx+' ',paradat)
        else:  ##  just use split(); assume already tokenized
            toks = rx.split()
        if paradat.wordonly:
            for t in toks:
                if t and t[0].isalnum():
                    toklist.append(t)  ##  only append if initial is alphanumeric
        else:
            toklist.extend(toks)  ##  add to growing list
        
    return  toklist


##  PUNKTRAN :
def punktran (punxlist=Punxlist):
    "creates a translation table from apos/quote/dash char codes."

    gofadict = {}  ##  table values = good old-fashioned ascii-type characters
    ck = 33  ##  default backup
    for cc in punxlist:
        if cc < 8:  ##  nice character
            ck = abs(cc)
            ##  gofdict[cc] = []
        else:
            gofadict[cc] = ck

    return  gofadict  ##  2b used by translate()


##  BASTATS :
def bastats (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = sum2
    statsack.std1 = statsack.std2 = math.sqrt(abs(sum2))
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  GEOMEAN :
def geomean (xvec):
    "computes geometric mean of vector of postive values."

    n = len(xvec)
    if n < 1:
        return  0.0
    sumlogs = 0.0
    for v in xvec:
        if v < Tiny:
            return  0.0  ##  if any zeroes, result is zero
        sumlogs += math.log(v)
    return  math.exp(sumlogs/n)



##  CORSUMS :
def corsums (x,y):
    "produces sums required for correlation coefficient, in a Sack()."

    cs = Sack(0)
    n = len(x)
    n = min(n,len(y))  ##  chops extra if unequal
    if n < 1:
        print("Problem in corsums!!")
        return cs
    sx = sy = 0.0
    sxsq = sysq = 0.0
    sumprods = 0.0
    for i in range(n):
        sx += x[i] ; sy += y[i]
        sxsq += x[i]**2 ; sysq += y[i]**2
        sumprods += x[i] * y[i]

    cs.n = float(n)
    cs.sumprods = sumprods
    cs.sx = sx
    cs.sy = sy
    cs.sxsq = sxsq
    cs.sysq = sysq
    return  cs


##  CORRVAL :
def corrval (cs,epsilon=1e-23):
    "computes correlation coefficient from sums etc. in Sack()."

    cnum = cs.sumprods - (cs.sx*cs.sy) / cs.n
    tx = cs.sxsq - cs.sx**2 / cs.n
    ty = cs.sysq - cs.sy**2 / cs.n
    ##  try:
        ##  dnom = math.sqrt(tx*ty)
    ##  except ValueError as xxxx:
        ##  print(xxxx)
        ##  print(tx,ty,cnum,cs.n)
        ##  print(cs.sx,cs.sy,cs.sxsq,cs.sysq)
        ##  dnom = 0
    ##  dnom = math.sqrt(abs(tx*ty))  ##  danger of rounding errors
    dnom = math.sqrt(max(0.0,tx*ty))
    if dnom > epsilon:
        return  cnum / dnom
    elif tx < epsilon and ty < epsilon:
        return  1.0  ##  both flat
    else:
        return  0.0  ##  null case
    ##  semi-safe, but vulnerable to rounding errors.


##  RANX :
def ranx (xvec):
    "returns rank vector from raw data vector, with ties given mean positions."

    nx = len(xvec)
    ##  xvec.sort()  ##  ascending
    vtab = {}
    for x in xvec:
        fx = vtab.get(x,0)
        vtab[x] = fx + 1
    vvec = list(vtab.items())
    vvec.sort()
    ranktab = {}
    p = 0
    ranksum = 0
    for v in vvec:
        vx = v[0] ; vf = v[1]
        rx = (ranksum*2+vf)*0.5-0.5  ##  plus 0.5 if lowest rank is 1
        ranksum += vf
        ##  print(p,vx,vf,rx)
        ranktab[vx] = rx
        p += 1

    rankvec = []
    for x in xvec:
        rankvec.append(ranktab[x])
    return  rankvec  ##  returning ranktab more useful ?


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  REQUEST :
def request (prompter,default='',brax="[]",punk=" : "):
    "user query with default response."

    if default==0 or default:  ##  numeric zero is valid, null isn't
        r = input(str(prompter) + ' ' + brax[0] + str(default) + brax[1] + punk)
        if not r:
            r = default
    else:  ##  just normal input
        r = input(str(prompter) + punk)

    return  str(r)  ##  returns a string


##  GETMODE :
def getmode (q,modelist):
    "Gets mode-code from user."

    print(q)  ##  user prompt
    mode = -99
    while mode < 0 or mode >= len(modelist):
        mc = 0
        for m in modelist:
            print(mc,m)
            mc = mc + 1
        i = input("Give option number: ").strip()
        if len(i) < 1:
            i = '0'  ##  blank -> zero
        if i:
            try:
                mode = eval(i)
            except:
                mode = -99
        else:
            mode = -99
    print("Mode",mode,"chosen :",modelist[mode])
    return  mode


##  DIVERSIM :
def diversim (pvec,epsilon=1e-23):
    "Simpson's index of diversity 4a frequency or probability vector."

    n = len(pvec)
    if n < 2:
        return  0.0
    psum = sum(pvec)
    if psum < epsilon:
        raise RuntimeError("improper negative freqs/probs given to diversim !!")
    ##  silent amendment (perilous?) :
    pv = [p/psum for p in pvec]
    ps = [p*p for p in pv]  ##  squares

    return  1 - sum(ps)

